#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "lua.h"
#include "lauxlib.h"
#include "lualib.h"

const int N = 13;
const int C_EASY = 2;
const int C_HARD = 4000;
const int T = 300;

#define CHECK(cond) ((cond) ? (void)0 : (void)(fprintf(stderr, "Check failed at %d: %s\n", __LINE__, #cond), exit(2)))

#define WRONG(...) (fprintf(stdout, "Wrong answer: " __VA_ARGS__), exit(1))

void readfile(const char *filename, long* size, char** content) {
  FILE *f = fopen(filename, "r");
  CHECK(f != NULL);
  CHECK(fseek(f, 0, SEEK_END) == 0);
  CHECK((*size = ftell(f)) != -1);
  *content = (char *)calloc(*size + 1, 1);
  CHECK(*content != NULL);
  CHECK(fseek(f, 0, SEEK_SET) == 0);
  CHECK(fread(*content, 1, *size, f) == *size);
  CHECK(!ferror(f));
  CHECK(fclose(f) == 0);
}

void check_pcall(lua_State* L, int nargs, int nresults, int errfunc) {
  int status = lua_pcall(L, nargs, nresults, errfunc);
  if (status != LUA_OK) {
    const char *err = lua_tostring(L, -1);
    WRONG("Lua error: %s\n", err ? err : "(error object is not a string)");
  }
}

int main(int argc, char** argv) {
  if (argc != 3) {
    fprintf(stderr, "usage: %s [easy|hard] program.lua\n", argv[0]);
    return 47;
  }

  srandom(47123);

  int C = argv[1][0] == 'e' ? C_EASY : C_HARD;

  long helper_size;
  char* helper_content;
  readfile("helper.lua", &helper_size, &helper_content);

  long program_size;
  char* program_content;
  readfile(argv[2], &program_size, &program_content);

  for (int t = 0; t < T; t++) {
    int colors[N];
    int hands[N];
    int answers[N];

    for (int i = 0; i < N; i++) {
      colors[i] = 1 + (random() % C);
    }

    lua_State* states[N];

    for (int i = 0; i < N; i++) {
      lua_State* L = states[i] = luaL_newstate();
      CHECK(L != NULL);

      luaL_openlibs(L);

      CHECK(luaL_loadbufferx(L, helper_content, helper_size, "=helper", "t") == LUA_OK);

      lua_pushinteger(L, N);
      lua_pushinteger(L, C);
      lua_pushinteger(L, i+1);
      lua_pushlstring(L, program_content, program_size);
      lua_createtable(L, N, 0);
      for (int j = 0; j < N; j++) {
        if (i != j) {
          lua_pushinteger(L, colors[j]);
          lua_seti(L, -2, j+1);
        }
      }

      check_pcall(L, 5, 2, 0);

      hands[i] = lua_toboolean(L, -2);
    }

    for (int i = 0; i < N; i++) {
      lua_State* L = states[i];
      CHECK(lua_gettop(L) == 2);

      lua_createtable(L, N, 0);
      for (int j = 0; j < N; j++) {
        lua_pushboolean(L, hands[j]);
        lua_seti(L, -2, j+1);
      }

      check_pcall(L, 1, 1, 0);

      CHECK(lua_isnumber(L, -1));
      answers[i] = lua_tointeger(L, -1);
    }

    int correct = 1;
    for (int i = 0; i < N; i++) {
      if (colors[i] != answers[i]) correct = 0;
    }

    if (!correct) WRONG("Guessed the wrong color.\n");

    for (int i = 0; i < N; i++) {
      lua_close(states[i]);
    }
  }

  fprintf(stdout, "OK\n");
  return 0;
}
